/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Sun Microsystems, Inc.
 * Portions created by the Initial Developer are Copyright (C) 1999
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include <ThreadManagerTests.h>

/*Sleep*/

TM_OJIAPITest(ThreadManager_Sleep_1) {
	GET_TM_FOR_TEST
	PRUint32 id = 0;
	class DummyThread : public BaseDummyThread {
	private:
		nsIThreadManager *tm;
	public:
		int awoken;
		DummyThread(nsIThreadManager *aTm) : tm(aTm), awoken(0) {}
		NS_METHOD Run() {
			tm->Sleep(0);
			awoken = 1;
			return NS_OK;
		}
	};

	DummyThread *newThread = new DummyThread(threadMgr);
	nsresult rc  = threadMgr->CreateThread(&id, (nsIRunnable*)newThread);
	if (NS_SUCCEEDED(rc)) {
		rc = threadMgr->Sleep(500);        
		if (NS_SUCCEEDED(rc) && newThread->awoken) //if it waked up -> pass
			return TestResult::PASS("Method should work OK.");
		return TestResult::FAIL("Sleep", rc);
	} 
	return TestResult::FAIL("Sleep", "Can't create new thread", rc);

}



TM_OJIAPITest(ThreadManager_Sleep_2) {
	GET_TM_FOR_TEST
	PRUint32 id = 0;
	class DummyThread : public BaseDummyThread {
	private:
		nsIThreadManager *tm;
	public:
		int isSleeping;
		DummyThread(nsIThreadManager *aTm) : tm(aTm), isSleeping(1) {} //UINT_MAX = 32-bit uint max value 
		NS_METHOD Run() {
			tm->Sleep(UINT_MAX);
			isSleeping = 0;
			return NS_OK;
		}
	};

	DummyThread *newThread = new DummyThread(threadMgr);
	nsresult rc  = threadMgr->CreateThread(&id, (nsIRunnable*)newThread);
	if (NS_SUCCEEDED(rc)) {
		rc = threadMgr->Sleep(500);        
		if (NS_SUCCEEDED(rc) && newThread->isSleeping)
			return TestResult::PASS("Method should work OK.");
		return TestResult::FAIL("Sleep", rc);
	} 
	return TestResult::FAIL("Sleep", "Can't create new thread", rc);

}
