/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla Communicator client code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Bill Law       law@netscape.com
 *   Jonathan Wilson <jonwil@tpgi.com.au>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include <windows.h>
#include "nsNativeAppSupportWin.h"

1 24 "MozillaSunbird.exe.manifest"

IDI_APPICON ICON "sunbird.ico"
IDI_APPLICATION ICON "sunbird.ico"

STRINGTABLE DISCARDABLE
BEGIN
#ifdef DEBUG
    IDS_STARTMENU_APPNAME,              "Sunbird Debug"
#else
    IDS_STARTMENU_APPNAME,              "Sunbird"
#endif
END

#ifdef MOZ_STATIC_BUILD

// XXX This code is copied from resource.h and widget.rc.  It's a work-around
// for the limitation that only one resource file can be used in an .exe.  We
// should develop a method, for static builds only, to combine multiple .rc
// files into a single .rc file, and then use that to build the single .res
// file for the .exe.

// We need this cursor for versions of Windows that don't support IDC_HAND,
// namely Win95 and NT4.
#define IDC_SELECTANCHOR                4100
#define IDC_GRAB                        4101
#define IDC_GRABBING                    4102
#define IDC_CELL                        4103
#define IDC_COPY                        4104
#define IDC_ALIAS                       4105
#define IDC_ZOOMIN                      4106
#define IDC_ZOOMOUT                     4107
#define IDC_COLRESIZE                   4108
#define IDC_ROWRESIZE                   4109
#define IDC_VERTICALTEXT                4110

IDC_SELECTANCHOR        CURSOR  DISCARDABLE     "..\\..\\widget\\src\\build\\res\\select.cur"
IDC_GRAB                CURSOR  DISCARDABLE     "..\\..\\widget\\src\\build\\res\\grab.cur"
IDC_GRABBING            CURSOR  DISCARDABLE     "..\\..\\widget\\src\\build\\res\\grabbing.cur"
IDC_CELL                CURSOR  DISCARDABLE     "..\\..\\widget\\src\\build\\res\\cell.cur"
IDC_COPY                CURSOR  DISCARDABLE     "..\\..\\widget\\src\\build\\res\\copy.cur"
IDC_ALIAS               CURSOR  DISCARDABLE     "..\\..\\widget\\src\\build\\res\\aliasb.cur"
IDC_ZOOMIN              CURSOR  DISCARDABLE     "..\\..\\widget\\src\\build\\res\\zoom_in.cur"
IDC_ZOOMOUT             CURSOR  DISCARDABLE     "..\\..\\widget\\src\\build\\res\\zoom_out.cur"
IDC_COLRESIZE           CURSOR  DISCARDABLE     "..\\..\\widget\\src\\build\\res\\col_resize.cur"
IDC_ROWRESIZE           CURSOR  DISCARDABLE     "..\\..\\widget\\src\\build\\res\\row_resize.cur"
IDC_VERTICALTEXT        CURSOR  DISCARDABLE     "..\\..\\widget\\src\\build\\res\\vertical_text.cur"

#endif
