.buttonBoxButton {
  -moz-appearance: none;
  border: 1px solid transparent;
  background-color: transparent;
  margin-left: 1px;
  margin-right: 1px;
  list-style-image: url("chrome://calendar/skin/Options.png")
}

#calendarGeneralButton   {-moz-image-region: rect(0px, 32px,  32px, 0px)}
#calendarAlarmButton   {-moz-image-region: rect(0px, 64px,  32px, 32px)}
#calendarPublishbutton  {-moz-image-region: rect(0px, 96px,  32px, 64px)}
#calendarMiscaButton {-moz-image-region: rect(0px, 128px, 32px, 96px)}
#calendarViewButton    {-moz-image-region: rect(0px, 160px, 32px, 128px)}
#calendarTZButton       {-moz-image-region: rect(0px, 192px, 32px, 160px)}
#calendarMiscbButton  {-moz-image-region: rect(0px, 224px, 32px, 192px)}

